/*
 * Copyright (C) 2004 Claudio "KLaN" Ciccani <klan82@cheapnet.it>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *
 *  written by Claudio Ciccani <klan82@cheapnet.it> with the contribute
 *  of Kristof Pelckmans <kristof.pelckmans@antwerpen.be>
 *
 */

#ifdef HAVE_CONFIG_H
# include <config.h>
#endif

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <errno.h>

#include "df_xine.h"


#define MSG_TIMEOUT   300000ll /* in microseconds */
#define STEP_SEEK      20000   /* in milliseconds */
#define STEP_COLORS      514
#define STEP_VOLUME        1


#ifdef CLAMP
 #undef CLAMP
#endif

#define CLAMP( n, min, max )  n = (n < min) ? min : ((n > max) ? max : n);



static const char *ratio_names[] =
{
     [0] = "auto",
     [1] = "Square",
     [2] = "4:3",
     [3] = "Anamorphic",
     [4] = "DVB"
};




static const char*
dfx_error( DFXCore *this )
{
     int err = xine_get_error( this->stream );

     switch (err)
     {
          case XINE_ERROR_NO_INPUT_PLUGIN:
               return "no input plugin";
          break;

          case XINE_ERROR_NO_DEMUX_PLUGIN:
               return "no demuxer plugin";
          break;

          case XINE_ERROR_DEMUX_FAILED:
               return "demux failed";
          break;

          case XINE_ERROR_MALFORMED_MRL:
               return "malformed mrl";
          break;

          default:
          break;
     }

     return "generic error";
}


static int
dfx_getpl( DFXCore *this )
{
     int pos = -1;
     int len = -1;
     int try =  5;

     while (try--)
     {
          if (xine_get_pos_length( this->stream,
                         NULL, &pos, &len ))
               break;

          usleep( 1000 );
     }

     if (pos < 0 || len < 1)
          return 0;

     this->ctl.pos = pos;
     this->ctl.len = len;

     return 1;
}


static int
dfx_control_play( DFXCore *this, int toggle )
{
     DFXVideoControl *ctl    = &this->ctl;
     DFXMedia        *media  = &this->media;
     int              width;
     int              height;
     int              status;
     
     if (toggle)
     {
          status = xine_get_status( this->stream );

          if (status == XINE_STATUS_PLAY)
          {
               DBUG( "stopping mrl '%s'\n", media->cur->mrl );
               ctl->status = XINE_STATUS_STOP;
               dfx_display_status( this, DDF_INFO );
               xine_stop( this->stream );
               usleep( 100 );
               return 1;
          }
     } else
     {
          xine_stop( this->stream );
          xine_close( this->stream );

          if (!media->cur)
          {
               DBUG( "playlist is empty\n" );
               return 0;
          }
          
          if (!xine_open( this->stream, media->cur->mrl ))
          {
               SAY( "skipping mrl '%s' [%s]\n",
                    media->cur->mrl, dfx_error( this ) );
               dfx_playlist_discard( this );
               return dfx_control_play( this, 0 );
          }

          width  = xine_get_stream_info( this->stream,
                         XINE_STREAM_INFO_VIDEO_WIDTH  );
          height = xine_get_stream_info( this->stream,
                         XINE_STREAM_INFO_VIDEO_HEIGHT );

          if (width > 0 && height > 0)
               dfx_context_update( this, width, height );
     }
          
     if (!xine_get_stream_info( this->stream,
                    XINE_STREAM_INFO_HAS_VIDEO ))
          dfx_context_set_gmode( this, DGM_ANIM );
     else
          dfx_context_set_gmode( this, DGM_VIDEO );

     if (!xine_play( this->stream, 0, 0 )) 
     {
          SAY( "skipping mrl '%s' [%s]\n",
               media->cur->mrl, dfx_error( this ) );
          dfx_playlist_discard( this );
          return dfx_control_play( this, 0 );
     }
     
     DBUG( "playing mrl '%s'\n", media->cur->mrl );

     usleep( 200 );
     
     return 1;
}


static void
dfx_control_pause( DFXCore *this )
{
     DFXVideoControl *ctl   = &this->ctl;
     int              speed;
     
     if (xine_get_status( this->stream )
                    != XINE_STATUS_PLAY)
          return;
     
     speed = xine_get_param( this->stream, XINE_PARAM_SPEED );

     if (speed != XINE_SPEED_PAUSE)
          speed = XINE_SPEED_PAUSE;
     else
          speed = XINE_SPEED_NORMAL;

     xine_set_param( this->stream, XINE_PARAM_SPEED, speed);

     usleep( 100 );

     dfx_getpl( this );
     
     ctl->speed = speed;
     dfx_display_status( this, DDF_INFO );
}


static void
dfx_control_seek( DFXCore *this, int param )
{
     DFXVideoControl *ctl = &this->ctl;
     
     if (!xine_get_stream_info( this->stream,
                    XINE_STREAM_INFO_SEEKABLE ))
          return;
     
     if (xine_get_status( this->stream )
                    != XINE_STATUS_PLAY)
          return;

     if (!dfx_getpl( this ))
          return;

     switch (param)
     {
          case DAP_MINUS:
          case DAP_PLUS:
               ctl->pos += param * STEP_SEEK; /* seek relative */
          break;

          default:
               ctl->pos  = (ctl->len * param) / 100; /* seek absolute */
          break;
     }

     if (ctl->pos > ctl->len)
          ctl->pos = ctl->len;
     else 
     if (ctl->pos < 0)
          ctl->pos = 0;

     xine_play( this->stream, 0, ctl->pos );

     usleep( 100 );

     dfx_display_status( this, DDF_BAR );
}


static void
dfx_control_speed( DFXCore *this, int param )
{
     DFXVideoControl *ctl   = &this->ctl;
     int              speed;

     if (xine_get_status( this->stream )
                    != XINE_STATUS_PLAY)
          return;

     speed = xine_get_param( this->stream, XINE_PARAM_SPEED );
     
     if (speed == XINE_SPEED_PAUSE)
          return;

     if (param < 0)
          speed = (speed + 1) >> 1;
     else
          speed = (speed < XINE_SPEED_FAST_4) ?
                  (speed << 1) : speed;

     DBUG( "setting speed to %i\n", speed );

     xine_set_param( this->stream, XINE_PARAM_SPEED, speed );

     ctl->speed = speed;
     dfx_display_status( this, DDF_INFO );
}


static void
dfx_control_getcsc( DFXCore *this, int csc )
{
     DFXVideoControl    *ctl = &this->ctl;
     DFXVideoContext    *ctx = &this->ctx;
     DFBColorAdjustment  adj;

     ctx->layer->GetColorAdjustment( ctx->layer, &adj );

     switch (csc)
     {
          case DAI_BRIGHTNESS:
          {
               if (ctx->caps & DLCAPS_BRIGHTNESS)
                    ctl->brightness = adj.brightness;
               else
                    ctl->brightness = xine_get_param( this->stream,
                                   XINE_PARAM_VO_BRIGHTNESS );
               dfx_display_bar( this, "Brightness",
                         ctl->brightness * 100 / 0xffff );
          }
          break;

          case DAI_CONTRAST:
          {
               if (ctx->caps & DLCAPS_CONTRAST)
                    ctl->contrast = adj.contrast;
               else
                    ctl->contrast = xine_get_param( this->stream,
                                   XINE_PARAM_VO_CONTRAST );
               dfx_display_bar( this, "Contrast",
                         ctl->contrast * 100 / 0xffff );
          }
          break;

          case DAI_SATURATION:
          {
               if (ctx->caps & DLCAPS_SATURATION)
                    ctl->saturation = adj.saturation;
               else
                    ctl->saturation = xine_get_param( this->stream,
                                   XINE_PARAM_VO_SATURATION );
               dfx_display_bar( this, "Saturation",
                         ctl->saturation * 100 / 0xffff );
          }
          break;

          case DAI_HUE:
          {
               if (ctx->caps & DLCAPS_HUE)
                    ctl->hue = adj.hue;
               else
                    ctl->hue = 0;
               dfx_display_bar( this, "Hue",
                         ctl->hue * 100 / 0xffff );
          }
          break;

          default:
          break;
     }
}


static void
dfx_control_setcsc( DFXCore *this, int csc )
{
     DFXVideoControl    *ctl = &this->ctl;
     DFXVideoContext    *ctx = &this->ctx;
     DFBColorAdjustment  adj =
     {
          .flags      = DCAF_NONE,
          .brightness = ctl->brightness,
          .contrast   = ctl->contrast,
          .saturation = ctl->saturation,
          .hue        = ctl->hue
     };

     switch (csc)
     {
          case DAI_BRIGHTNESS:
          {
               if (ctx->caps & DLCAPS_BRIGHTNESS)
                    adj.flags |= DCAF_BRIGHTNESS;
               else
                    xine_set_param( this->stream,
                              XINE_PARAM_VO_BRIGHTNESS,
                              ctl->brightness );
               dfx_display_bar( this, "Brightness",
                         ctl->brightness * 100 / 0xffff );
          }
          break;

          case DAI_CONTRAST:
          {
               if (ctx->caps & DLCAPS_CONTRAST)
                    adj.flags |= DCAF_CONTRAST;
               else
                    xine_set_param( this->stream,
                              XINE_PARAM_VO_CONTRAST,
                              ctl->contrast );
               dfx_display_bar( this, "Contrast",
                         ctl->contrast * 100 / 0xffff );
          }
          break;

          case DAI_SATURATION:
          {
               if (ctx->caps & DLCAPS_SATURATION)
                    adj.flags |= DCAF_SATURATION;
               else
                    xine_set_param( this->stream,
                              XINE_PARAM_VO_SATURATION,
                              ctl->saturation );
               dfx_display_bar( this, "Saturation",
                         ctl->saturation * 100 / 0xffff );
          }
          break;

          case DAI_HUE:
          {
               if (ctx->caps & DLCAPS_HUE)
                    adj.flags |= DCAF_HUE;
               dfx_display_bar( this, "Hue",
                         ctl->hue * 100 / 0xffff );
          }
          break;

          default:
          break;
     }

     if (adj.flags)
          ctx->layer->SetColorAdjustment( ctx->layer, &adj );
}


static void
dfx_control_ratio( DFXCore *this, int param )
{
     DFXVideoControl *ctl = &this->ctl;
     DFXVideoContext *ctx = &this->ctx;

     switch (param)
     {
          case DAP_RESET:
               ctl->ratio = 0.0;
          break;

          case DAP_RATIO_SQUARE:
               ctl->ratio = (double) ctx->video.width /
                         (double) ctx->video.height;
          break;

          case DAP_RATIO_43:
               ctl->ratio = 4.0 / 3.0;
          break;

          case DAP_RATIO_169:
               ctl->ratio = 16.0 / 9.0;
          break;

          case DAP_RATIO_DVB:
               ctl->ratio = 2.0 / 1.0;
          break;

          default:
          return;
     }

     ctx->update = 1;
     
     if (param != ctl->ratio_id)
     {
          ctl->ratio_id = param;
          dfx_display_text( this, "Ratio: %s", 
                    ratio_names[ctl->ratio_id] );
     }
}


static void
dfx_control_audio( DFXCore *this, int set, int param )
{
     char buffer[32];
     int  channel;

     if (set)
     {
          channel = xine_get_param( this->stream,
                         XINE_PARAM_AUDIO_CHANNEL_LOGICAL );

          if (param == DAP_RESET)
               channel  = -1;
          else
               channel += param;

          xine_set_param( this->stream,
                    XINE_PARAM_AUDIO_CHANNEL_LOGICAL, channel );
          usleep( 100 );
     }
     
     channel = xine_get_param( this->stream,
                    XINE_PARAM_AUDIO_CHANNEL_LOGICAL );

     if (!xine_get_audio_lang( this->stream, channel, &buffer[0] ))
          sprintf( buffer, "auto" );

     dfx_display_text( this, "Audio Channel: %s", buffer );
}


static void
dfx_control_spu( DFXCore *this, int set, int param )
{
     char buffer[32];
     int  channel;

     if (set)
     {
          channel = xine_get_param( this->stream, 
                         XINE_PARAM_SPU_CHANNEL );

          if (param == DAP_RESET)
               channel  = -1;
          else
               channel += param;

          if (channel < -1)
               channel = -1; /* avoid turning off spu (will disable OSD) */

          xine_set_param( this->stream,
                    XINE_PARAM_SPU_CHANNEL, channel );
     }
     
     channel = xine_get_param( this->stream, XINE_PARAM_SPU_CHANNEL );

     if (!xine_get_spu_lang( this->stream, channel, &buffer[0] ))
          sprintf( buffer, "auto" );

     dfx_display_text( this, "Subtitles Channel: %s", buffer );
}



static void*
dfx_control_handler( void *data )
{
     DFXCore         *this    = (DFXCore*) data;
     DFXVideoControl *ctl     = &this->ctl;
     int              notice;
     int              actid;
     char             param;
     int              monitor = 0;

     pthread_mutex_lock( &ctl->mutex );
     
     if (!dfx_control_play( this, 0 ))
          this->quit = 2;

     while (!this->quit)
     {
          struct timeval  now;
          struct timespec timeout;

          gettimeofday( &now, NULL );

          timeout.tv_sec  = now.tv_sec;
          timeout.tv_nsec = (now.tv_usec + 500000) * 1000;
          
          if (pthread_cond_timedwait( &ctl->cond, &ctl->mutex,
                         &timeout ) == ETIMEDOUT)
          {
               if (monitor && dfx_getpl( this ))
               {
                    ctl->status = xine_get_status( this->stream );
                    ctl->speed  = xine_get_param( this->stream,
                                        XINE_PARAM_SPEED );
                    dfx_display_status( this, DDF_BOTH );
               }

               continue;
          }
          
          gettimeofday( &now, NULL );

          if(((now.tv_sec * 1000000ll + now.tv_usec) -
              (ctl->time.tv_sec * 1000000ll + ctl->time.tv_usec)) > MSG_TIMEOUT)
          {
               SAY( "discarding message %#x, older than %lli micros.\n",
                    ctl->magic, MSG_TIMEOUT );
               continue;
          }
          
          notice = DFX_NOTICE( ctl->magic );
          actid  = DFX_ACTION( ctl->magic );
          param  = DFX_PARAM( ctl->magic );

          if (notice == DAN_ACQUIRE)
          {
               switch (actid)
               {
                    case DAI_TOGGLE_PLAY:
                    case DAI_PLAY:
                    case DAI_TOGGLE_PAUSE:
                    case DAI_SEEK:
                    case DAI_SPEED:
                    {
                         dfx_getpl( this );
                         ctl->status = xine_get_status( this->stream );
                         ctl->speed  = xine_get_param( this->stream,
                                             XINE_PARAM_SPEED );
                         dfx_display_status( this, DDF_BOTH );
                         monitor = 1;
                    }
                    break;

                    case DAI_VOLUME:
                    {
                         ctl->volume = xine_get_param( this->stream,
                                        XINE_PARAM_AUDIO_VOLUME );
                         dfx_display_bar( this, "Volume", ctl->volume );
                    }
                    break;

                    case DAI_BRIGHTNESS:
                    case DAI_CONTRAST:
                    case DAI_SATURATION:
                    case DAI_HUE:
                         dfx_control_getcsc( this, actid );
                    break;

                    case DAI_RATIO:
                         dfx_display_text( this, "Ratio: %s",
                                   ratio_names[ctl->ratio_id] );
                    break;

                    case DAI_AUDIO_CHANNEL:
                         dfx_control_audio( this, 0, 0 );
                    break;

                    case DAI_SPU_CHANNEL:
                         dfx_control_spu( this, 0, 0 );
                    break;

                    default:
                    break;
               }
               
          } else
          if (notice == DAN_RELEASE)
          {
               switch (actid)
               {
                    case DAI_TOGGLE_PLAY:
                    case DAI_PLAY:
                    case DAI_TOGGLE_PAUSE:
                    case DAI_SEEK:
                    case DAI_SPEED:
                         monitor = 0;
                    break;

                    default:
                    break;
               }

               dfx_display_hide( this, 0 );
               
          } else
          {
               switch (actid)
               {
                    case DAI_QUIT:
                         this->quit = 1;
                    break;

                    case DAI_TOGGLE_PLAY:
                    {
                         if (!dfx_control_play( this, 1))
                              this->quit = 2;
                    }
                    break;

                    case DAI_PLAY:
                    {
                         dfx_playlist_skip( this, param );
                         if (!dfx_control_play( this, 0 ))
                              this->quit = 2;
                    }
                    break;

                    case DAI_TOGGLE_PAUSE:
                         dfx_control_pause( this );
                    break;

                    case DAI_SEEK:
                         dfx_control_seek( this, param );
                    break;

                    case DAI_SPEED:
                         dfx_control_speed( this, param );
                    break;

                    case DAI_VOLUME:
                    {
                         ctl->volume += param * STEP_VOLUME;
                         CLAMP( ctl->volume, 0, 100 );
                         xine_set_param( this->stream,
                                   XINE_PARAM_AUDIO_VOLUME,
                                   ctl->volume );
                         dfx_display_bar( this, "Volume", ctl->volume );
                    }
                    break;

                    case DAI_BRIGHTNESS:
                    {
                         ctl->brightness += param * STEP_COLORS;
                         CLAMP( ctl->brightness, 0, 0xffff );
                         dfx_control_setcsc( this, DAI_BRIGHTNESS );
                    }
                    break;

                    case DAI_CONTRAST:
                    {
                         ctl->contrast += param * STEP_COLORS;
                         CLAMP( ctl->contrast, 0, 0xffff );
                         dfx_control_setcsc( this, DAI_CONTRAST );
                    }
                    break;

                    case DAI_SATURATION:
                    {
                         ctl->saturation += param * STEP_COLORS;
                         CLAMP( ctl->saturation, 0, 0xffff );
                         dfx_control_setcsc( this, DAI_SATURATION );
                    }
                    break;

                    case DAI_HUE:
                    {
                         ctl->hue += param * STEP_COLORS;
                         CLAMP( ctl->hue, 0, 0xffff );
                         dfx_control_setcsc( this, DAI_HUE );
                    }
                    break;

                    case DAI_RATIO:
                         dfx_control_ratio( this, param );
                    break;

                    case DAI_AUDIO_CHANNEL:
                         dfx_control_audio( this, 1, param );
                    break;

                    case DAI_SPU_CHANNEL:
                         dfx_control_spu( this, 1, param );
                    break;

                    case DAI_OSD_HIDE:
                         dfx_display_hide( this, 1 );
                    break;
                    
                    default:
                    break;
               }
          }
     }

     pthread_mutex_unlock( &ctl->mutex );

     return NULL;
}


static void
dfx_event_listner( void *data, const xine_event_t *event )
{
     DFXCore *this = (DFXCore*) data;

     switch (event->type)
     {
          case XINE_EVENT_UI_PLAYBACK_FINISHED:
          {
               DFXMedia *media = &this->media;
               int       magic;

               if (media->cur)
               {
                    SAY( "playback finished for mrl '%s'\n", media->cur->mrl );
                    
                    /* end of playlist and no more loops */
                    if (media->cur == media->last && media->repeat > 0 &&
                        ++media->repeated >= media->repeat)
                         break;
                    
                    magic = DFX_MAGIC( 0, 0, DAI_PLAY, DAP_PLUS );
                    dfx_control_action( this, magic );
               }
          }
          break;

          case XINE_EVENT_FRAME_FORMAT_CHANGE:
          {
               xine_format_change_data_t *data;

               data = (xine_format_change_data_t*) event->data;
               if (data)
                    dfx_context_update( this, 
                         data->width, data->height );
          }
          break;

          default:
          break;
     }
}


void
dfx_control_start( DFXCore *this )
{
     DFXVideoControl *ctl   = &this->ctl;
     pthread_t       thread;
     pthread_mutex_t mutex  = PTHREAD_MUTEX_INITIALIZER;
     pthread_cond_t  cond   = PTHREAD_COND_INITIALIZER;

     pthread_mutex_init( &mutex, NULL );
     pthread_cond_init( &cond, NULL );

     if (pthread_create( &thread, NULL,
               dfx_control_handler, (void*) this ))
     {
          pthread_cond_destroy( &cond );
          pthread_mutex_destroy( &mutex );
          FATAL( "couldn't start control thread\n" );
     }

     ctl->thread = thread;
     ctl->mutex  = mutex;
     ctl->cond   = cond;

     ctl->queue = xine_event_new_queue( this->stream );
     if (ctl->queue)
          xine_event_create_listener_thread( ctl->queue,
                    dfx_event_listner, (void*) this );
}


void
dfx_control_action( DFXCore *this, int magic )
{
     DFXVideoControl *ctl = &this->ctl;

     if (DFX_EVENT( magic ))
     {
          xine_event_t evt = 
          {
               .data        = (void*) &evt,
               .data_length = sizeof( xine_event_t ),
               .type        = DFX_EVENT( magic ),
               .stream      = this->stream
          };
          
          gettimeofday( &evt.tv, NULL );
          xine_event_send( this->stream, &evt );
     } else
     {
          struct timeval now;

          gettimeofday( &now, NULL );
          
          pthread_mutex_lock( &ctl->mutex );
          
          ctl->magic = magic;
          ctl->time  = now;
          pthread_cond_signal( &ctl->cond );

          pthread_mutex_unlock( &ctl->mutex );
     }
}


void
dfx_control_release( DFXCore *this )
{
     DFXVideoControl *ctl = &this->ctl;
     void            *ret;

     if (ctl->thread)
     {
          pthread_cancel( ctl->thread );
          pthread_join( ctl->thread, &ret );
          pthread_cond_destroy( &ctl->cond );
          pthread_mutex_destroy( &ctl->mutex );
     }

     dfx_display_release( this );

     if (ctl->queue)
          xine_event_dispose_queue( ctl->queue );
}


